/* SPDX-License-Identifier: MIT
 * Small Data Compression Library
 * Copyright (c) 2020 Kingizor */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sd_internal.h"


/* Error reporting functions */
static const char *sd_error_msg = "";

const char *sd_get_error (void) {
    return sd_error_msg;
}
void sd_set_error (const char *error) {
    sd_error_msg = error;
}



/* File/Buffer handling */
static int check_size_comp (size_t input_size) {
    if (input_size > 0x10000) {
        sd_set_error("Input is too large to fit in a 16-bit bank. (n > 65536)");
        return 1;
    }
    if (input_size < 64) {
        sd_set_error("Input is too small. (n < 64)");
        return 1;
    }
    return 0;
}
static int check_size_decomp (size_t input_size) {
    if (input_size > 16777216) {
        sd_set_error("Input file is too large.");
        return 1;
    }
    if (input_size < 6) {
        sd_set_error("Input is too small. (n < 6)");
        return 1;
    }
    return 0;
}
static int check_input_mem (unsigned char *input, size_t input_size, int mode_comp) {
    if (input == NULL) {
        sd_set_error("Input pointer is NULL.");
        return 1;
    }
    if (mode_comp) {
        if (check_size_comp(input_size))
            return 1;
    }
    else {
        if (check_size_decomp(input_size))
            return 1;
    }
    return 0;
}

static int open_input_file (const char *fn, unsigned char **input, size_t *input_size, int ofs, int mode_comp) {
    FILE *f;
    long input_length;

    f = fopen(fn, "rb");
    if (f == NULL) {
        sd_set_error("Failed to open input file.");
        return 1;
    }

    if ((fseek(f, 0,   SEEK_END)  == -1)
    || ((input_length = ftell(f)) == -1)
    ||  (fseek(f, ofs, SEEK_SET)  == -1)) {
        sd_set_error("Failed to seek the input file.");
        goto error;
    }

    if (input_length <= ofs) {
        sd_set_error("File offset is larger than the input file.");
        goto error;
    }

    if (mode_comp) {
        if (check_size_comp(input_length))
            goto error;
        *input_size = input_length;
    }
    else {
        if (check_size_decomp(input_length))
            goto error;
        if (ofs > input_length) {
            sd_set_error("Supplied decompression offset is larger than the input size.");
            goto error;
        }
        if (ofs < 0) {
            sd_set_error("Supplied decompresison offset is negative.");
            goto error;
        }
        *input_size = input_length - ofs;
        if (*input_size > 0x20028) {
            *input_size = 0x20028;
        }
        if (*input_size < 64) {
            sd_set_error("File size minus offset is too small.");
        }
    }

    *input = malloc(*input_size);
    if (*input == NULL) {
        sd_set_error("Failed to allocate memory for input buffer.");
        goto error;
    }

    if (fread(*input, 1, *input_size, f) != (size_t)*input_size) {
        sd_set_error("Failed to read from input file.");
        free(*input);
        goto error;
    }
    return 0;
error:
    if (f != NULL)
        fclose(f);
    return 1;
}

static int write_output_file (const char *fn, unsigned char *output, size_t output_size) {

    FILE *f;

    f = fopen(fn, "wb");
    if (f == NULL) {
        sd_set_error("Failed to open file for writing.");
        return 1;
    }

    if (fwrite(output, 1, output_size, f) != output_size) {
        sd_set_error("An error occurred while writing the output file.");
        fclose(f);
        return 1;
    }

    fclose(f);
    return 0;
}
static int open_output_buffer (unsigned char **output, size_t output_size) {
    *output = calloc(output_size, 1);
    if (*output == NULL) {
        sd_set_error("Failed to allocate memory for output buffer.");
        return 1;
    }
    return 0;
}



/* Compression handlers */

int sd_compress_mem_to_mem (
    unsigned char **output,
    size_t *outsize,
    unsigned char *input,
    size_t insize
) {
    struct SD_COMPRESSOR sd;
    memset(&sd, 0, sizeof(struct SD_COMPRESSOR));

    if (check_input_mem(input, insize, 1))
        return 1;

    sd.input  = input;
    sd.insize = insize;

    sd.outsize = 0x10000;
    if (open_output_buffer(&sd.output, sd.outsize))
        return 1;
    if (sd_compress(&sd)) {
        free(sd.output);
        return 1;
    }
    *output  = sd.output;
    *outsize = sd.outsize;
    return 0;
}

int sd_compress_file_to_mem (
    unsigned char **output,
    size_t *outsize,
    const char *file_in
) {
    struct SD_COMPRESSOR sd;
    memset(&sd, 0, sizeof(struct SD_COMPRESSOR));

    if (open_input_file(file_in, &sd.input, &sd.insize, 0, 1))
        goto error;

    sd.outsize = 0x10000;
    if (open_output_buffer(&sd.output, sd.outsize))
        goto error;

    if (sd_compress(&sd))
        goto error;

    free(sd.input);
    *output  = sd.output;
    *outsize = sd.outsize;
    return 0;
error:
    free(sd.input);
    free(sd.output);
    return 1;
}


int sd_compress_mem_to_file (
    const char *file_out,
    unsigned char *input,
    size_t insize
) {
    unsigned char *output;
    size_t outsize;

    if (sd_compress_mem_to_mem(&output, &outsize, input, insize))
        return 1;
    if (write_output_file(file_out, output, outsize)) {
        free(output);
        return 1;
    }
    free(output);
    return 0;
}

int sd_compress_file_to_file (
    const char *file_out,
    const char *file_in
) {
    unsigned char *output;
    size_t outsize;

    if (sd_compress_file_to_mem(&output, &outsize, file_in))
        return 1;
    if (write_output_file(file_out, output, outsize)) {
        free(output);
        return 1;
    }
    free(output);
    return 0;
}



/* Decompression handlers */

int sd_decompress_mem_to_mem (
    unsigned char **output,
    size_t *outsize,
    unsigned char *input,
    size_t insize
) {
    struct SD_COMPRESSOR sd;
    memset(&sd, 0, sizeof(struct SD_COMPRESSOR));

    if (check_input_mem(input, insize, 0))
        return 1;

    sd.input  = input;
    sd.insize = insize;
    sd.outsize = 0x10000;

    if (open_output_buffer(&sd.output, sd.outsize))
        return 1;

    if (sd_decompress(&sd)) {
        free(sd.output);
        return 1;
    }
    *output  = sd.output;
    *outsize = sd.outsize;
    return 0;
}
int sd_decompress_file_to_mem (
    unsigned char **output,
    size_t *outsize,
    const char *file_in,
    int position
) {
    struct SD_COMPRESSOR sd;
    memset(&sd, 0, sizeof(struct SD_COMPRESSOR));

    if (open_input_file(file_in, &sd.input, &sd.insize, position, 0))
        goto error;

    sd.outsize = 0x10000;
    if (open_output_buffer(&sd.output, sd.outsize))
        goto error;
    if (sd_decompress(&sd))
        goto error;

    free(sd.input);
    *output  = sd.output;
    *outsize = sd.outsize;
    return 0;
error:
    free(sd.input);
    free(sd.output);
    return 1;
}

int sd_decompress_mem_to_file (
    const char *file_out,
    unsigned char *input,
    size_t insize
) {
    unsigned char *output;
    size_t outsize;

    if (sd_decompress_mem_to_mem(&output, &outsize, input, insize))
        return 1;
    if (write_output_file(file_out, output, outsize)) {
        free(output);
        return 1;
    }
    free(output);
    return 0;
}
int sd_decompress_file_to_file (
    const char *file_out,
    const char *file_in,
    size_t position
) {
    unsigned char *output;
    size_t outsize;

    if (sd_decompress_file_to_mem(&output, &outsize, file_in, position))
        return 1;
    if (write_output_file(file_out, output, outsize)) {
        free(output);
        return 1;
    }
    free(output);
    return 0;
}


