/* SPDX-License-Identifier: MIT
 * Small Data Compression Library
 * Copyright (c) 2020 Kingizor */

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include "sd_internal.h"

jmp_buf error_handler;

static unsigned char rb (struct SD_COMPRESSOR *sd, size_t addr) {
    if (addr >= sd->insize) {
        sd_set_error("Tried to read out of bounds.");
        longjmp(error_handler, 1);
    }
    return sd->input[addr];
}
static unsigned short rw (struct SD_COMPRESSOR *sd, size_t addr) {
    return rb(sd, addr) | (rb(sd, addr+1) << 8);
}

/* read n bits from input */
static unsigned short bits (struct SD_COMPRESSOR *sd, int count) {
    unsigned val = 0;
    while (count--) {
        unsigned char bit = !!(rb(sd, sd->inpos) & (1 << (sd->bitpos ^ 7)));
        val |= bit << count;
        if (++sd->bitpos == 8) {
            sd->inpos++;
            sd->bitpos = 0;
        }
    }
    return val;
}

/* modify word (all writes are rmw) */
static void mw (struct SD_COMPRESSOR *sd, size_t addr, int val) {
    addr <<= 1;
    if (addr+1 >= sd->outsize) {
        sd_set_error("Attempted to write out of bounds.");
        longjmp(error_handler, 1);
    }
    sd->output[addr  ] |= val;
    sd->output[addr+1] |= val >> 8;
}


/* These routines (four variants) determine the upper 6 bits */
static void sub_decompress (struct SD_COMPRESSOR *sd, int mode) {

    size_t addr = 0;
    unsigned shift, count_size, val_size;
    if (mode == 3) {
          val_size = 3;
        count_size = 4;
        shift = 10;
    }
    else {
          val_size = 1;
        count_size = 6;
        shift = 13 + mode;
    }

    for (;;) {
        unsigned char  loop  =        bits(sd, 1);
        unsigned short val   =        bits(sd, val_size) << shift;
        unsigned char  count = loop ? bits(sd, count_size) : 1;

        if (!count)
            break;
        while (count--)
            mw(sd, addr++, val);
    }

}

/* This routine determines values to be placed in the low 10 bits */
static void main_decompress (struct SD_COMPRESSOR *sd) {

    size_t addr = 0;

    for (;;) {
        unsigned char  mode = bits(sd,  2);
        unsigned short val  = bits(sd, 10);
        unsigned short count;

        if (!mode) { /* write single value once */
            count = 1;
        }
        else if (mode == 1) { /* write single value 1-63 times  */
            count = bits(sd, 6);
            if (!count) /* Quit if zero */
                break;
        }
        else { /* write incrementing or decrementing value 1-15 times */
            count = bits(sd, 4);

            /* These cases don't have the exit condition,
               so the game would write 65536 values in succession */
            if (!count) {
                fprintf(stderr,
                    "Encountered an exit condition in a case without "
                    "an exit check."
                );
                longjmp(error_handler, 1);
            }
        }

        while (count--) {
            mw(sd, addr++, val);
            if (mode == 2)
                val++;
            else if (mode == 3)
                val--;
            val &= 0x3FF;
        }
    }
}

int sd_decompress (struct SD_COMPRESSOR *sd) {

    int i, subs;

    if (setjmp(error_handler))
        return 1;

    /* first byte indicates which subs to call */
    subs     = rb(sd, sd->inpos + 0) & 7;

    /* next word is the output size in words */
    sd->outsize = rw(sd, sd->inpos + 1) << 1;

    sd->inpos  += 3;

    /* first three subs are optional */
    for (i = 0; i < 3; i++)
        if (subs & (1 << i))
            sub_decompress(sd, i); /* {2,4,8}000 */

    /* fourth sub and the main routine are mandatory */
    sub_decompress(sd, 3); /* 1C00 */
    main_decompress(sd);   /* 03FF */
    return 0;
}

